/*
 * hash.c - mIRC-like cryptographic hashing driver
 * Copyright (c) 2011, D.C. van Moolenbroek, XISE <dc@xise.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "hash.h"
#include "mirc.h"

static unsigned int scan_val(LPTSTR *str)
{
  unsigned int val;
  LPTSTR p;

  for (p = *str; *p == ' '; p++);

  for (val = 0; *p >= '0' && *p <= '9'; p++)
   val = val * 10 + *p - '0';

  while (*p && *p != ' ') p++;

  *str = p;

  return val;
}

#define MAX_BVAR_BYTES (MIRC_LINE_LENGTH / 4)

static BOOL hash_binvar(LPTSTR name, hash_func *hf, void *ctx, HWND hwnd)
{
  LPTSTR data, p;
  DWORD index, bytes;
  unsigned char buf[MAX_BVAR_BYTES];
  BOOL res;

  if (lstrlen(name) + 32 >= MIRC_LINE_LENGTH) return FALSE;

  if ((data = mirc_open()) == NULL) return FALSE;

  for (index = 1; ; index += MAX_BVAR_BYTES) {
    wsprintf(data, TEXT("$bvar(%s,%u,%u)"), name, index, MAX_BVAR_BYTES);

    if (!(res = mirc_query(hwnd))) break;

    p = data;

    for (bytes = 0; bytes < MAX_BVAR_BYTES && *p; bytes++)
     buf[bytes] = (unsigned char) scan_val(&p);

    if (bytes > 0) hf->update(ctx, buf, bytes);

    if (bytes < MAX_BVAR_BYTES) break;
  }

  mirc_close();

  return res;
}

static BOOL hash_file(LPTSTR name, hash_func *hf, void *ctx)
{
  HANDLE handle;
  unsigned char buf[4096];
  DWORD bytes;
  BOOL res;

  handle = CreateFile(name, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE |
   FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

  if (handle == INVALID_HANDLE_VALUE) return FALSE;

  while ((res = ReadFile(handle, buf, sizeof(buf), &bytes, NULL)) == TRUE &&
   bytes > 0) {
    hf->update(ctx, buf, bytes);
  }

  CloseHandle(handle);

  return res;
}

int hash(LPTSTR data, hash_func *hf, unsigned char *digest, unsigned int size,
 void *ctx, HWND hwnd)
{
  unsigned int i, mode;
  LPTSTR p;

  hf->init(ctx);

  p = data;
  mode = scan_val(&p);

  switch (mode) {
  case 0: /* Plain text */
    if (*p == ' ') p++;

    hf->update(ctx, (unsigned char *) p, lstrlen(p) * sizeof(*p));

    break;

  case 1: /* Binary variable */
    while (*p == ' ') p++;

    if (*p != '&') return 1;

    if (!hash_binvar(p, hf, ctx, hwnd)) return 1;

    break;

  case 2: /* File */
    while (*p == ' ') p++;

    if (!hash_file(p, hf, ctx)) return 1;

    break;

  default: /* Unknown */
    return 1;
  }

  hf->final(ctx, digest);

  for (i = 0; i < size; i++)
   wsprintf(&data[i * 2], TEXT("%02x"), digest[i]);

  return 3;
}
